/*
 *
 * This file is a part of gcd - gcd is an ncurses based audio CD player.
 *
 * Copyright (C) 1998  Geza Herman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Geza Herman
 * <hg211@ural2.hszk.bme.hu>
 *
 */

#include "display.h"
#include "gcd.h"
#include "error.h"
#include <malloc.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#define  SETCOLOR( x )	attrset( x )

#define	 _CF	SETCOLOR( C_CYAN );
#define  _CN	standend();
#define  _CG	SETCOLOR( C_GREEN );
#define  _CY    SETCOLOR( C_YELLOW );
#define  _CB    SETCOLOR( C_BLUE | A_BOLD );
#define  _CR	SETCOLOR( C_RED );
#define  _CW	SETCOLOR( C_WHITE );
#define  _CC	SETCOLOR( C_CYAN | A_BOLD );

static int	maxx, maxy;
static int	tracks_h;
static int	extend_y, extend_h;
static int	ptracks_y, ptracks_h;

static void	set_colors()
{
	init_pair( 1, COLOR_WHITE, COLOR_BLACK );
	init_pair( 2, COLOR_YELLOW, COLOR_BLACK );
	init_pair( 3, COLOR_CYAN, COLOR_BLACK );
	init_pair( 4, COLOR_GREEN, COLOR_BLACK );
	init_pair( 5, COLOR_RED, COLOR_BLACK );
	init_pair( 6, COLOR_BLUE, COLOR_BLACK );
	init_pair( 7, COLOR_BLUE, COLOR_CYAN );
	init_pair( 8, COLOR_YELLOW, COLOR_CYAN );
}

static char *strndup( char *s, int n )
{
	int	l = strlen( s );
	int	nl = l > n ? n : l;
	char   *ns = malloc( nl + 1 );
	
	strncpy( ns, s, nl );
	ns[ nl ] = '\0';
	return( ns ); 
}

void 	init_display()
{
	initscr();
	getmaxyx( stdscr, maxy, maxx );
	if( maxx < 80 || maxy < 24 ) {
	    endwin();
	    fprintf( stderr, "Error: Screen size is too small.\n" );
	    exit( ERR_OPNSCR );
	}    

	if( has_colors() ) {
	    start_color();
	    set_colors();
	} else {
	    endwin();
	    fprintf( stderr, "Error: This terminal doesn't support colors.\n" );
	    exit( ERR_OPNSCR );
	}        

	tracks_h = ( maxy - 5 ) / 2;
	ptracks_h = ( maxy - 5 ) * 2 / 5;
	if( ptracks_h < 8 ) ptracks_h = 8;
	extend_h = maxy - 5 - tracks_h - ptracks_h;
	extend_y = 3 + tracks_h;
	ptracks_y = extend_y + extend_h + 1;
	
	cbreak();
	noecho();
	nonl();
	leaveok( stdscr, TRUE );
	intrflush( stdscr, FALSE );
	keypad( stdscr, TRUE );
	scrollok( stdscr, FALSE );
	halfdelay( 5 );
}

void 	done_display()
{
	clear();
	refresh();
	endwin();
}

static void	print_padded( int y, int x, char *s, int n )
{
        char   *ps = malloc( n + 1 );
	int	w = strlen( s ) > n ? n : strlen( s );
        
        ps[ n ] = '\0';

        memcpy( ps , s, w );
        memset( ps + w, ' ', n - w );
        mvprintw( y, x, ps );
        free( ps );	
}

void	init_screen()
{
	_CW mvprintw( 0, 0, "gcd "VERSION" " );
	_CF mvaddch( 0, 9, ACS_ULCORNER );
	
	    print_cdtitle( NULL, FALSE );
	
	_CF mvaddch( 0, maxx - 10, ACS_URCORNER );
	_CW mvprintw( 0, maxx - 9, " by  Geza" ); 
	
	_CF
	    move( 1, 0 );
	    hline( ACS_HLINE, 9 );
	    mvaddch( 1, 9, ACS_BTEE );
	    hline( ACS_HLINE, maxx - 20 );
	    mvaddch( 1, maxx - 10, ACS_BTEE );
	    hline( ACS_HLINE, 9 );
	    mvaddch( extend_y - 1, 0, ACS_HLINE );
	    
	    
	_CG mvprintw( extend_y - 1, 1, " Extended information " );
	
	print_wh_ext( NULL );
	
	_CF mvaddch( ptracks_y - 1, 0, ACS_HLINE );
	_CG mvprintw( ptracks_y - 1, 1, " Programmed tracks " );
	
	_CF 
	    hline( ACS_HLINE, maxx - 45 );  
	    mvaddch( ptracks_y - 1, maxx - 25, ACS_TTEE );
	
	_CG mvprintw( ptracks_y - 1, maxx - 24, " CD Information " );
	    
	_CF    
	    hline( ACS_HLINE, 8 );   
	    move( ptracks_y, maxx - 25 );
	    vline( ACS_VLINE, maxy - ptracks_y - 1 );   
	    mvaddch( maxy - 1, 0, ACS_HLINE );
	
	_CB mvprintw( maxy - 1, 1, " Error: " );
	
	print_errormsg( NULL );
	
	_CC
	    mvprintw( ptracks_y    , maxx - 23, "State:" );
	    mvprintw( ptracks_y + 1, maxx - 23, "Playmode:" );
	    mvprintw( ptracks_y + 2, maxx - 23, "Total CD Time:" );
	    mvprintw( ptracks_y + 3, maxx - 23, "CD Time:" );
	    mvprintw( ptracks_y + 4, maxx - 23, "Track Time:" );
	    mvprintw( ptracks_y + 5, maxx - 23, "Left volume:" );
	    mvprintw( ptracks_y + 6, maxx - 23, "Right volume:" );
	    mvprintw( ptracks_y + 7, maxx - 23, "Disc ID:" );
		        
}

void	print_discID( unsigned ID )
{
	char	buf[ 9 ];
	
	_CY
	if( ID != 0 ) {
	    sprintf( buf, "%08x", ID );
	    mvprintw( ptracks_y + 7, maxx - 14, buf );
	} else {
	    mvprintw( ptracks_y + 7, maxx - 14, "        " );
	}        
}

void	print_volume( u_char left, u_char right )
{
        _CY
        mvprintw( ptracks_y + 5, maxx - 10, "%d%%  ", left * 100 / 255 );
        mvprintw( ptracks_y + 6, maxx - 9, "%d%%  ", right * 100 / 255 );
}

void	print_state( int s )
{
	_CY
	switch( s ) { 
	    case ST_STOPPED: mvprintw( ptracks_y, maxx - 16, "Stopped" ); break;
	    case ST_PAUSED : mvprintw( ptracks_y, maxx - 16, "Paused " ); break;
	    case ST_PLAYING: mvprintw( ptracks_y, maxx - 16, "Playing" ); break;
	    case ST_UNKNOWN: mvprintw( ptracks_y, maxx - 16, "Unknown" ); break;
	    case ST_NODISK : mvprintw( ptracks_y, maxx - 16, "No Disk" ); break;
	}
}

void	print_cdtime( int min, int sec, char prefix )
{
	_CY
	if( min != -1 ) {
	    mvprintw( ptracks_y + 3, maxx - 15, "%c%02d:%02d", prefix, min, sec );
	} else {
	    mvprintw( ptracks_y + 3, maxx - 15, " --:--" );
	}
}

void	print_tracktime( int min, int sec, char prefix )
{
	_CY
	if( min != -1 ) {
            mvprintw( ptracks_y + 4, maxx - 12, "%c%02d:%02d", prefix, min, sec );
	} else {
	    mvprintw( ptracks_y + 4, maxx - 12, " --:--" );
	}
}

void	print_cdtotal( int min, int sec, int frame )
{
	_CY
	if( min != -1 ) {
	    mvprintw( ptracks_y + 2, maxx - 8, "%02d:%02d.%02d", min, sec, frame * 4 / 3 );
	} else {
	    mvprintw( ptracks_y + 2, maxx - 8, "--:--.--" );
	}
}

void	print_playmode( int pm, int r )
{
	int p;
	
	_CY
	switch( pm ) {
	   case -1:         mvprintw( ptracks_y + 1, maxx - 13, "-"       ); p = 1; break;
	   case PM_NORMAL:  mvprintw( ptracks_y + 1, maxx - 13, "Normal"  ); p = 6; break;
	   case PM_SHUFFLE: mvprintw( ptracks_y + 1, maxx - 13, "Shuffle" ); p = 7; break;
	   case PM_PROGRAM: mvprintw( ptracks_y + 1, maxx - 13, "Program" ); p = 7; break;
	   case PM_INTRO:   mvprintw( ptracks_y + 1, maxx - 13, "Intro"   ); p = 5; break;
	}
	if( r && pm != -1 ) {
	    mvprintw( ptracks_y + 1, maxx - 13 + p, " Rep" );
	    p += 4;
	}
	move( ptracks_y + 1, maxx - 13 + p );
	hline( ' ', 13 - p );     
}

void	init_input( struct input_inf *i )
{
	i->len = -1;
}

int	input( int y, int x, int width, char *s, int maxlen, struct input_inf *i )
{
        int	ch;
	int 	quit = FALSE;

#define len	 ( i->len )
#define cursor	 ( i->cursor )
#define	first	 ( i->first )
#define updated	 ( i->updated )
#define modified ( i->modified )

	if( len == -1 ) {
	    len = strlen( s );
	    cursor = len;
	    first = len - width < 0 ? 0 : len - width;
	    updated = TRUE;
	    modified = FALSE;
	} 

	leaveok( stdscr, FALSE );
	do {
	    if( updated ) {
	        mvaddch( y, x - 1, first == 0 ? ' ' : ACS_LARROW );
	        mvaddch( y, x + width, first + width >= len ? ' ' : ACS_RARROW );
	        print_padded( y, x, s + first, width );
	        updated = FALSE;
	    }    
            move( y, x + cursor - first );

	    switch( ch = getch() ) {
	        case KEY_LEFT: if( cursor > 0 ) { cursor--; updated = TRUE; } break;
	        case KEY_RIGHT: if( cursor < len ) { cursor++; updated = TRUE; } break;
	        case KEY_HOME: cursor = 0; updated = TRUE; break;
	        case KEY_END: cursor = len; updated = TRUE; break;
	        case MKEY_ENTER: quit = IN_ACCEPT; break;
	        case MKEY_ESC: quit = IN_CANCEL; break;
	        case ERR: quit = IN_AGAIN; break;
	        case KEY_BACKSPACE: 
	            if( cursor > 0 && len > 0 ) {
	                memmove( s + cursor - 1, s + cursor, strlen( s + cursor ) + 1 );
	                cursor--;
	                len--;
	                updated = modified = TRUE;
	            }
	            break;
	        case KEY_DC:
	            if( cursor < len ) {
	                memmove( s + cursor, s + cursor + 1, strlen( s + cursor + 1 ) + 1 );
	                len--;
	                updated = modified = TRUE;
	            }
	            break;
	        default:
	            if( ( ( ch >= 0x20 && ch <= 0x7e ) || ( ch >= 0xa0 && ch <= 0xff ) ) && len < maxlen ) {
			memmove( s + cursor + 1, s + cursor, strlen( s + cursor ) + 1 );
			s[ cursor ] = ch;
			cursor++;
			len++;
			updated = modified = TRUE;
	            }
	            break;    
	    }

	    if( first > cursor ) { first = cursor; updated = TRUE; }
	    if( first + width < cursor ) { first = cursor - width; updated = TRUE; }
	} while( !quit );    

#undef  modified
#undef  updated
#undef	cursor
#undef	len
#undef	first

        if( quit != IN_AGAIN ) leaveok( stdscr, TRUE );
        return( quit );
}

static char    *nextline( char *s )
{
	int	found = FALSE;
	
	while( *s && !found ) {
	    if( *s == '\\' ) {
	        if( *( s + 1 ) == 'n' ) found = TRUE;
	        s += 2;
	    } else {
	        s++;
	    }
	}                    
	return( *s == '\0' ? NULL : s );
}

static char    *linestart( char *b, char *s )    
{
	char  *cl;
	
	while( b != NULL && b <= s ) {
	    cl = b;
	    b = nextline( b );
	}    
	return( cl );    
}

static int	linelength( char *s )
{
	int	l = 0;
	int	f = FALSE;
	
	while( !f && *s ) {
	    l++;
	    if( *s == '\\' ) {
	        if( *( s + 1 ) == 'n' ) {
	            f = TRUE;
	            l--;
	        }    
	        s += 2;
	    } else {
	        s++;
	    }
	}       
	return( l );     
}

static char    *passchars( char *s, int n )
{
	int	i;
	
	for( i = 0; i < n; i++ ) s += ( *s == '\\' ? 2 : 1 );
	return( s );             
}  

void	print_extline( int y, int x, int w, int x0, char *s )
{
	char   *p = malloc( w + 1 );
	char   *e = nextline( s );
	char   *cp = p;	
	
	e = ( e == NULL ? s + strlen( s ) : e - 2 );
	memset( p, ' ', w );
	p[ w ] = '\0';    
	
	while( x0 && s < e && cp < p + w ) {
            if( *s == '\\' ) {
                if( *( s + 1 ) == '\\' ) x0--;
	        s += 2;
	    } else {
	        s++;
	        x0--;
	    }        		            
	}
	while( s < e && cp < p + w ) {
	    if( *s == '\\' ) {
	        if( *( s + 1 ) == '\\' ) *cp++ = '\\';
	        s += 2;
	    } else {
	        *cp++ = *s++;
	    }
	}                
	mvprintw( y, x, p );
	
	free( p );	
}

void	print_ext( int y0, int x0, char *s )
{
	int	i;

	SETCOLOR( C_CYAN );
	while( y0 > 0 ) {
	    if( s != NULL ) s = nextline( s );
	    y0--;
	}
	
	for( i = 0; i < extend_h; i++ ) {   
	    if( s != NULL ) {
	        print_extline( extend_y + i, 0, maxx, x0, s );
	        s = nextline( s );
	    } else {
	        move( extend_y + i, 0 );
	        hline( ' ', maxx );
	    }
	}             	
}                                       

void	print_wh_ext( char *w )
{
	if( w == NULL ) {
	    _CF
	    move( extend_y - 1, 23 ); 
	    hline( ACS_HLINE, maxx - 23 );
	} else {
	    char       *b = strndup( w, maxx - 27 ); 	
	    int		l = strlen( b ) + 3;
	    
	    _CG
	    mvprintw( extend_y - 1, 23, "(%s) ", b );
	    free( b );
	    
	    _CF
	    move( extend_y - 1, 23 + l );
	    hline( ACS_HLINE, maxx - 23 - l );
	}    
}

int	input_ext( char *s, int maxlen, struct input_inf *i )
{
	int	quit = FALSE;
	int	ch;

#define len	 ( i->len )
#define cursorx	 ( i->cursor )
#define cursory  ( i->cursory )
#define x0	 ( i->first )
#define y0	 ( i->y0 )
#define cl	 ( i->cl )
#define	c	 ( i->c )
#define updated	 ( i->updated )
#define modified ( i->modified )
	
	if( len == -1 ) {
	    len = strlen( s );
	    cursorx = cursory = x0 = y0 = 0;
	    cl = c = s;
	    updated = TRUE;
	    modified = FALSE;
	} 

	leaveok( stdscr, FALSE );
	do {
	    if( updated ) {
	        print_ext( y0, x0, s );
	        updated = FALSE;
	    }    
            move( extend_y + cursory - y0, cursorx - x0 );

	    switch( ch = getch() ) {
	        case MKEY_ESC: quit = IN_CANCEL; break;
	        case MKEY_TAB: quit = IN_ACCEPT; break;
	        case ERR: quit = IN_AGAIN; break;
	        case MKEY_ENTER:
	            if( len < maxlen - 1 ) {
	                memmove( c + 2, c, strlen( c ) + 1 );
	                *c++ = '\\';
	                *c++ = 'n';
	                cl = c;
	                cursorx = 0;
	                cursory++;
	                len += 2;
	                updated = modified = TRUE;
	            }
	            break;
	        case KEY_UP:
	            if( cursory ) {
	                int	l;
	                
	                cursory--;
	                cl = linestart( s, cl - 2 );
	                l = linelength( cl );
	                if( cursorx > l ) cursorx = l;
	                c = passchars( cl, cursorx );
	                updated = TRUE;
	            }
	            break;
	        case KEY_DOWN: {
	            char       *e = c + linelength( c );
	            int		l;
	            
	            if( *e == '\\' ) {
	                cursory++;
	                cl = e + 2;
	                l = linelength( cl );
	                if( cursorx > l ) cursorx = l;
	                c = passchars( cl, cursorx );
	                updated = TRUE;
	            }    
		    break;
	        }
		case KEY_HOME:
		    c = cl;
		    cursorx = 0;
		    updated = TRUE;
		    break;
		case KEY_END: {
		    int	d = linelength( c );
		    
		    c += d;
		    cursorx += d;
		    updated = TRUE;
		    break;
		}    
	        case KEY_RIGHT:
	            if( *c ) {
	                if( *c == '\\' && *( c + 1 ) == 'n' ) {
	                    cursorx = 0;
	                    cursory++;
	                    cl = c = c + 2;
	                } else {
	                    cursorx++;
	                    c = passchars( c, 1 );
	                }
	                updated = TRUE;    
	            }    
	            break;            
	        case KEY_LEFT:
	        case KEY_BACKSPACE:
	            if( cursorx == 0 ) {
	                if( cursory > 0 ) {
	                    cursory--;
	                    c -= 2;
	                    cl = linestart( s, c );
	                    cursorx = linelength( cl );
	                    updated = TRUE;
	                } else {
	                    break;
	                }    
	            } else {
	                cursorx--;
	                c = passchars( cl, cursorx );
	                updated = TRUE;
	            }
	            if( ch == KEY_LEFT ) break;
	        case KEY_DC:
	            if( *c ) {
	                char    *nc = passchars( c, 1 );
	                
	                memmove( c, nc, strlen( nc ) + 1 );
	                len -= nc - c;
	                updated = modified = TRUE;
	            }
	            break;    
	        default:
	            if( ( ( ch >= 0x20 && ch <= 0x7e ) || ( ch >= 0xa0 && ch <= 0xff ) ) && len < maxlen ) {
			if( ch == '\\' ) {
			    memmove( c + 2, c, strlen( c ) + 1 );
			    *c++ = '\\';
			    *c++ = '\\';
			    len += 2;
			} else {    
			    memmove( c + 1, c, strlen( c ) + 1 );
       			    *c++ = ch;
			    len++;
			}
			cursorx++;    
			updated = modified = TRUE;
	            }
	            break;    
	    }
	    
	    if( x0 > cursorx ) { x0 = cursorx; updated = TRUE; }		
	    if( x0 + maxx < cursorx ) { x0 = cursorx - maxx; updated = TRUE; }
	    if( y0 > cursory ) { y0 = cursory; updated = TRUE; }
	    if( y0 + extend_h - 1 < cursory ) { y0 = cursory - extend_h + 1; updated = TRUE; }
	    
	} while( !quit );

#undef  modified
#undef  updated
#undef	c
#undef	cl
#undef	y0
#undef	x0
#undef	cursory
#undef	cursorx
#undef	len

        if( quit != IN_AGAIN ) leaveok( stdscr, TRUE );
	return( quit );
}

void	print_errormsg( char *msg )
{
	int	l;
	int	x;
	
	if( msg == NULL ) {
	    _CB mvprintw( maxy - 1, 9, "no error " );
	    l = 8;
	} else {    
	    _CR mvprintw( maxy - 1, 9, "%s ", msg );
	    l = strlen( msg );
	}    
	
	_CF 
	    move( maxy - 1, l + 10 );
	    hline( ACS_HLINE, maxx - l - 35 );
	    if( l + 9 < maxx - 25 ) mvaddch( maxy - 1, maxx - 25, ACS_BTEE );
	    x = l + 9 < maxx - 25 ? maxx - 24 : l + 10;
	    move( maxy - 1, x );
	    hline( ACS_HLINE, maxx - x );
}

int	cdtitle_max()
{
	return( maxx - 22 );
}

void	print_cdtitle( char *t, int s )
{
	char   *n;
	int	l = 0;
	
	SETCOLOR( s ? C_CURRHIGH : C_CURRENT );
	if( t == NULL ) {
	    if( s ) { 
	        mvprintw( 0, 10, "  " );
	        l = 2;
	    }    
	} else {    
	    n = strndup( t, maxx - 22 );
	    mvprintw( 0, 10, " %s ", n );
	    l = strlen( n ) + 2;
	    free( n ); 
	}
	move( 0, l + 10 );
	_CF hline( ACS_HLINE, maxx - l - 20 );   
}

int	title_y()
{
	return( 2 );
}

int	title_x()
{
	return( 4 );
}

int	title_maxx()
{
	return( maxx - 23 );
}

int	title_maxy()
{
	return( tracks_h );
}

int	ext_y()
{
	return( extend_y );
}

int	ext_h()
{
	return( extend_h );
}

int	ext_w()
{
	return( maxx );
}

int	prog_maxy()
{
	return( ptracks_h );
}

int	prog_maxx()
{
	return( maxx - 26 );
}

void	set_color( u_char type, int tr, int p_track, int t_curr )
{
	SETCOLOR( C_BLUE | ( type ? A_BOLD : 0 ) );
	if( tr == p_track ) SETCOLOR( C_CURRENT );
	if( tr == t_curr ) SETCOLOR( type ? ( tr == p_track ? C_CURRHIGH : C_HIGHLIGHT ) : C_HIGHLIGHT_B );
}

void	print_title( int y, int tr, u_char type, char *t, int t_curr, int p_track, int min, int sec, int frame, int lmin, int lsec, int lframe, int ignore )
{
	set_color( type, tr, p_track, t_curr );
	mvprintw( y + 2, 0, "%02d. ", tr + 1 );
	if( t == NULL || ignore ) {
	    move( y + 2, 4 );
	    hline( ' ', maxx - 23 );
	} else {
	    print_padded( y + 2, 4, t, maxx - 23 );
	}
	mvprintw( y + 2, maxx - 19, "  %02d:%02d.%02d %02d:%02d.%02d", min, sec, frame * 4 / 3, lmin, lsec, lframe * 4 / 3 );
}                                         

void	delete_title( int y )
{
	_CN
	move( y + 2, 0 );
	hline( ' ', maxx );
}                      

void	print_prog( int y, int tr, char *t, int p_curr, int pp_track )
{
        SETCOLOR( tr == pp_track ? C_CURRENT : C_BLUE | A_BOLD );
        if( tr == p_curr ) SETCOLOR( tr == pp_track ? C_CURRHIGH : C_HIGHLIGHT );
        if( t == NULL ) {
            move( ptracks_y + y, 0 );
            hline( ' ', maxx - 26 );
        } else {
            print_padded( ptracks_y + y, 0, t, maxx - 26 );
        }         
}           

void	delete_prog( int y )
{
	_CN
	move( ptracks_y + y, 0 );
	hline( ' ', maxx - 26 );
}